module countrateerror_mod
!ttttttttttttttttttttttttttttttttttttttttttttttttttttttttt
!  VNY 21.11.2013 : simlified the routines for the rigorous 
! error-barcomputation
!tttttttttttttttttttttttttttttttttttttttttttttttttttttttt
!  of the error bars for the coincidence-loss corrected count rates
!  is added:
!
!  cLossErrBars(frameTime, timeDel, observedCountRate, verbFlag, deadFrac)
!
! The function returns the rms-error estimated with the use of the rigourous algorithm
! for error-bars (S.Rosen, 2004). 
!
! Input parameters:
! ----------------
! frameTime - real(kind=double): frame time in seconds
! timeDel - real: time of the observation, in seconds (it could be a bin-time or the e
! xposure time)
! deadFrac - real(kind=double): dead fraction in units of the fraction of the frame time
! observedCountRate - real : observed count rate (which is affecetd by the coincidence losses)
!
! Output parameters:
!-------------------
! trueCountRate - real: the estimated count rate (corrected for the coincidence losses); 
! this value can be used for checking purporses only
!
!
    use dal
    use cal
    use param
    use errorHandling
    use sasdal
    use omconstants_mod
    use omfunctions_mod


    use omsourceparam_mod
    implicit none

    public computesimonerrors
    public computesimonerrors2
       
    logical :: initialisedCLossErrBars = .false.
    integer(int16), private, parameter :: maxSize = 100
  

    real(double), private :: cLossFrameTime, cLossDeadFrac
 
    real(single), private :: cLossExposure
contains


!*****************************************************************

!***********************************************************
!
!
!**********************************************************
! 2013-11-21 re-shaping the routine by explicitely using the formula (7) 
! from Simon&Paul's 2008 paper
  subroutine ComputeSimonErrors(nsources,                        & 
       corrCountRate,                   &
       corrCountRateErr,                &
       corrBackgroundCountRate,         & 
       exposure,                        &
       deadFrac,                        &
       frameTime)
    integer(int32) :: nsources
    real(single), pointer, dimension(:) :: corrCountRate, corrCountRateErr
    real(single), pointer, dimension(:) :: corrBackgroundCountRate
    real(single), intent(in) :: exposure
       
    real(double), intent(in) :: deadFrac, frameTime
    real(single) :: corrBackgroundCountRateErr

    integer(int16) :: i
    real(single) :: totRate, totRateErr

    real(single) :: cminRate
    

    real :: rmsArraySimple(20000)


    integer :: maxCounts, nAux, iloss
    real(kind=double) :: maxCountsPerFrame
    real :: sim_frameTime, sim_timeBin, trueCountRate
    real :: maxLoss, rmsValue
    
 
    real(double) :: x

    real(single) :: nframes

    if (nsources .lt. 1) return

    clossexposure = 1.0
 
    cLossFrameTime = frameTime
    cLossDeadFrac = deadFrac

    !write(*,*) "L577 routine ComputeSimonsErrors ... "

    if (frameTime .gt. 0.0) then
       maxCountsPerFrame=0.99/frameTime
    else
       maxCountsPerFrame=200.0
    endif
    maxCounts=6*maxCountsPerFrame
    !write(*,*) "L596 maxCounts=", maxCounts
    if (maxCounts .gt. 2000) maxCounts=2000
    !write(*,*) "L588 nsources=", nsources, " maxSize=", maxSize, " maxCounts=", maxCounts

    sim_timeBin= exposure
    sim_frameTime=frameTime

    nframes=sim_timeBin/frameTime
    !write(*,*) "L609 nframes=", nframes
    
    do i=1,maxCounts*10
       trueCountRate=real(i)
       trueCountRate=trueCountRate/10.0

       x=trueCountRate
              
       call ComputeRmsValueSimple(frameTime, deadFrac, sim_timeBin, x, rmsValue)
       rmsArraySimple(i)=rmsValue
       
    enddo


    maxLoss=0.0
    !write(*,*) "L612 =====================================================+ maxCounts=", maxCounts
    do i=1,maxCounts
       trueCountRate=real(i)/10.0       
       !write(*,*) "L615 i=", i, " trueCountRate=", trueCountRate, " rmsSimple=",rmsArraySimple(i) 
    enddo ! i
    !write(*,*) "L612 =====================================================+ maxCounts=", maxCounts
 
    nAux=int(maxLoss)

    cminRate = 5.0
    !cminRate = 3.0

    cminRate = max(cminRate, minval(corrCountRate + corrBackgroundCountRate, &
         (corrCountRate+corrBackgroundCountRate) .gt. 0.0))
    
    !write(*,*) "L707 cminRate=", cminRate
    
       
    do i = 0, nsources - 1
       totRate = corrCountRate(i) + corrBackgroundCountRate(i)
       !write(*,*) ".......... totRate=", totRate
       
       if (totRate .lt. cminRate) then
          
          totRateErr=corrCountRateErr(i)
          !write(*,*) "L129 i=", i, " corrCountRate=", corrCountRate(i), &
          !     " corrBackgroundCountRate=", corrBackgroundCountRate(i), & 
          !     " totRate=", totRate, " totRateErr=", totRateErr, "  ... skipped "
       else
          !***********************************************************
          ! The error is picked up from the initiated array of errors 
          !***********************************************************         
          !iloss=int(totRate+0.55)
          iloss=int(totRate*10.0+0.55)
          
          if (iloss .gt. maxCounts*10) iLoss=maxCounts*10

          if (iloss .gt. 0) then 
          !if (iloss .gt. 1) then           
             totRateErr=rmsArraySimple(iloss)
             !totRateErr=rmsArraySimple(iloss-1)
          else
             totRateErr=corrCountRateErr(i)
          endif
          
          !write(*,*) "L696 i=", i, " iloss=", iloss, " rmsArraySimple=", &
          !     rmsArraySimple(iloss), " corrCountRate=", corrCountRate(i), &
          !     " corrCountRateErr=", corrCountRateErr(i), & 
          !     " totRate=", totRate, " totRateErr=", totRateErr
          
          !*******************************************************
          ! The error in the background  count-rate
          !*******************************************************
          !corrBackgroundCountRateErr = sqrt(corrBackgroundCountRate(i)) * sqrt(exposure) / exposure
          !corrCountRateErr(i) = sqrt(totRateErr * totRateErr + &
          !     2.0 * corrBackgroundCountRateErr * corrBackgroundCountRateErr)

          corrCountRateErr(i) = totRateErr

          
          !write(*, *) "L707    totRate=", totRate,  &
          !     " corrCountRateErr=",   corrCountRateErr(i)          
       endif
    end do
    !write(*,*) "L711 ================================"

end subroutine ComputeSimonErrors

!***********************************************************
!
!
!******************************************************************
! 2013-12-02 computing Simon errors using directly the formula (7)
!******************************************************************
  subroutine ComputeSimonErrors2(nsources,                        &
       rawCountRate,                   & 
       rawCountRateErr,                &
       rawBackgroundCountRate,         & 
       corrCountRate,                   & 
       corrCountRateErr,                &
       corrBackgroundCountRate,         & 
       exposure,                        &
       deadFrac,                        &
       frameTime)
    integer(int32) :: nsources
    real(single), pointer, dimension(:) :: rawCountRate, rawCountRateErr
    real(single), pointer, dimension(:) :: rawBackgroundCountRate
    real(single), pointer, dimension(:) :: corrCountRate, corrCountRateErr
    real(single), pointer, dimension(:) :: corrBackgroundCountRate
    real(single), intent(in) :: exposure
       
    real(double), intent(in) :: deadFrac, frameTime
    real(single) :: corrBackgroundCountRateErr

    integer(int16) :: i
    real(single) :: totRate, totRateErr

    real(single) :: cminRate
    

    real :: rmsArraySimple(20000)


    integer :: maxCounts, nAux, iloss
    real(kind=double) :: maxCountsPerFrame
    real :: sim_frameTime, sim_timeBin, trueCountRate
    real :: maxLoss, rmsValue
    
 
    real(double) :: x

    real(single) :: nframes
    real(single) :: sqrtError

    real(single) :: countsObserved
    real(single) :: nf_co, term1, alpha
    

    if (nsources .lt. 1) return

    clossexposure = 1.0
 
    cLossFrameTime = frameTime
    cLossDeadFrac = deadFrac
    alpha=1.0-deadFrac
    if (alpha .lt. 0.5) then
       alpha=0.5
    endif
    !write(*,*) "L247 routine ComputeSimonsErrors2 ... "

    if (frameTime .gt. 0.0) then
       maxCountsPerFrame=0.99/frameTime
    else
       maxCountsPerFrame=200.0
    endif
    maxCounts=6*maxCountsPerFrame
    !write(*,*) "L255 maxCounts=", maxCounts
    if (maxCounts .gt. 2000) maxCounts=2000
    !write(*,*) "L257 nsources=", nsources, " maxSize=", maxSize, " maxCounts=", maxCounts
    !write(*,*) "L269 deadFrac=", deadFrac, " alpha=", alpha

    sim_timeBin= exposure
    sim_frameTime=frameTime

    nframes=sim_timeBin/frameTime
    !write(*,*) "L264 nframes=", nframes
    
 !   do i=1,maxCounts*10
 !      trueCountRate=real(i)
 !      trueCountRate=trueCountRate/10.0
!
!       x=trueCountRate
!              
!       call ComputeRmsValueSimple(frameTime, deadFrac, sim_timeBin, x, rmsValue)
!       rmsArraySimple(i)=rmsValue
!       
!    enddo


!    maxLoss=0.0
!    write(*,*) "L279 =====================================================+ maxCounts=", maxCounts
!    do i=1,maxCounts
!       trueCountRate=real(i)/10.0       
!       write(*,*) "L282 i=", i, " trueCountRate=", trueCountRate, " rmsSimple=",rmsArraySimple(i) 
!    enddo ! i
!    write(*,*) "L284 =====================================================+ maxCounts=", maxCounts
! 
!    nAux=int(maxLoss)

    cminRate = 5.0
    !cminRate = 3.0

    cminRate = max(cminRate, minval(corrCountRate + corrBackgroundCountRate, &
         (corrCountRate+corrBackgroundCountRate) .gt. 0.0))
    
    !write(*,*) "L294 cminRate=", cminRate
    
       
    do i = 0, nsources - 1
       !totRate = corrCountRate(i) + corrBackgroundCountRate(i)
       totRate = (rawCountRate(i) + rawBackgroundCountRate(i))*exposure
       sqrtError=sqrt(totRate)
       !write(*,*) ".........i=", i
       !write(*,*) "L316 raw=", rawCountRate(i), " rawBkg=",  rawBackgroundCountRate(i), &
       !     " totRate=", totRate, " sqrtError=", sqrtError
       
       if (rawCountRate(i) .lt. cminRate) then
          
          totRateErr=corrCountRateErr(i)
          !write(*,*) "L322 i=", i, " corrCountRate=", corrCountRate(i), &
          !     " corrBackgroundCountRate=", corrBackgroundCountRate(i), & 
          !     " totRate=", totRate, " totRateErr=", totRateErr, "  ... skipped "
       else
          !***********************************************************
          ! The error is calculated by using Simon's formula 
          !***********************************************************
          if ((rawCountRate(i) + rawBackgroundCountRate(i))*frameTime .gt.0.97) then
             !write(*,*) "L330 i=", i, " rate exceeds 0.97 c/f"
             
             !corrCountRateErr(i)=nullValue
          else
             totRateErr=sqrtError
             countsObserved=totRate
             nf_co=nframes-countsObserved
             term1=totRateErr/0.1
             if (nf_co .gt. 0.1) then
                term1=totRateErr/nf_co
             endif
             !write(*,*) "L336 i=", i, " nframes=", nframes, " nf_co=", nf_co, " term1=", term1
             term1=dlog(1.0_double+dble(term1))
             totRateErr=nframes*term1/alpha
             
             !ttttttttttttttttttttttttttttttttttt
             
             !write(*,*) "L342 term1=", term1, " totRateErr=", totRateErr
             if (exposure .gt. 1.0) then
                totRateErr=totRateErr/exposure
             endif
             
             corrCountRateErr(i) = totRateErr
             
             !write(*, *) "L346    totRate=", totRate,  &
             !     " corrCountRateErr=",   corrCountRateErr(i)          
          endif
       endif
    end do
    !write(*,*) "L351 ================================"

end subroutine ComputeSimonErrors2

!---------------------------------------------------------------------------------
! Simplified the routine using Simon's coincidence loss test routine
! 2013-11-21 vny
! The simplified algorithm for calculating the CL-corrected count rate
! errors is based on the formula (7) from Kuin&Rosen paper MNRAS 383, 383 (2008).
! It calculates the error based on the true count rate which should be 
! estimated beforehand from the observed count rates by using CAL CL-routine.
! The formula is slightly transformed, so the formula used here is:
! sigma_counts = N_frames * ln{1+sqrt[(exp(trueCountsPerFrame -1)/N_frames]}/(1-deadFrac)
! sigma_countrate=sigma_counts/exposure
!---------------------------------------------------------------------------------
subroutine ComputeRmsValueSimple(frameTime, deadFrac, exposure, trueCountrate, rmsValue)
    real(double), intent(in) :: frameTime ! frame time 
    real(double), intent(in) :: deadFrac ! dead-time fraction
    real(single), intent(in) :: exposure ! total exposure
    real(double), intent(in) :: truecountrate ! [c/s] incident count rate (CL-corrected count rate)
  
    real(single), intent(out) :: rmsValue
 
    real(double) :: nFrames
    real(double) :: rms
    real(double) :: alpha

    real(double) :: trueCountsPerFrame

    rms = 0.0_double
    if (frameTime .gt. 0.0_double) then
       nFrames = dble(exposure) / frameTime
    else
       ! assume the frame time to be the standard 11 ms
       nFrames =  dble(exposure) /0.011_double
    endif

    ! dead-time correction factor:
    alpha=1.0_double-deadFrac
    if (alpha .lt. 0.5_double) then
       alpha=1.0_double
    endif

    trueCountsPerFrame=trueCountrate*frameTime
    if (nFrames .gt. 1.0_double) then       
       rms=1.0_double+sqrt((dexp(trueCountsPerFrame)-1.0_double)/nFrames)
    else
       rms=1.0_double+sqrt(dexp(trueCountsPerFrame)-1.0_double)
    endif
    rms=dlog(rms)*nFrames/alpha

    if (exposure .gt. 1.0) then
       rmsValue=real(rms)/exposure
    else
       rmsValue=real(rms)
    endif 
end subroutine computeRmsValueSimple

!*************************

end module countrateerror_mod

